@@ -15,8 +15,7 @@ from .losses import MNLLLoss
 from .losses import log1pMSELoss
 
 from .performance import pearson_corr
-from .performance import multinomial_log_probs
-from .performance import compute_performance_metrics
+from .performance import calculate_performance_measures
 
 torch.backends.cudnn.benchmark = True
 
@@ -72,31 +71,33 @@ class BPNet(torch.nn.Module):
 		The number of dilated residual layers to include in the model.
 		Default is 8.
 
-	n_profile_outputs: int, optional
+	n_outputs: int, optional
 		The number of profile outputs from the model. Generally either 1 or 2 
 		depending on if the data is unstranded or stranded. Default is 2.
 
-	n_count_outputs: int, optional
-		The number of count outputs from the model. Generally just 1, but
-		can be larger than 1 when multiple experiments are being modeled
-		at the same time. Default is 1.
-
 	alpha: float, optional
 		The weight to put on the count loss.
 
+	name: str or None, optional
+		The name to save the model to during training.
+
 	trimming: int or None, optional
 		The amount to trim from both sides of the input window to get the
 		output window. This value is removed from both sides, so the total
 		number of positions removed is 2*trimming.
 	"""
 
-	def __init__(self, n_filters=64, n_layers=8, n_outputs=2, alpha=1, 
-		trimming=None):
+	def __init__(self, n_filters=64, n_layers=8, n_outputs=2, 
+		n_control_tracks=2, alpha=1, profile_output_bias=True, 
+		count_output_bias=True, name=None, trimming=None):
 		super(BPNet, self).__init__()
 		self.n_filters = n_filters
 		self.n_layers = n_layers
 		self.n_outputs = n_outputs
+		self.n_control_tracks = n_control_tracks
+
 		self.alpha = alpha
+		self.name = name or "bpnet.{}.{}".format(n_filters, n_layers)
 		self.trimming = trimming or 2 ** n_layers
 
 		self.iconv = torch.nn.Conv1d(4, n_filters, kernel_size=21, padding=10)
@@ -106,56 +107,82 @@ class BPNet(torch.nn.Module):
 				dilation=2**i) for i in range(1, self.n_layers+1)
 		])
 
-		self.fconv = torch.nn.Conv1d(n_filters+2, n_outputs, kernel_size=75, 
-			padding=37)
+		self.fconv = torch.nn.Conv1d(n_filters+n_control_tracks, n_outputs, kernel_size=75, 
+			padding=37, bias=profile_output_bias)
 		
-		self.relu = torch.nn.ReLU()
-		self.logsoftmax = torch.nn.LogSoftmax(dim=-1)
+		n_count_control = 1 if n_control_tracks > 0 else 0
+		self.linear = torch.nn.Linear(n_filters+n_count_control, 1, 
+			bias=count_output_bias)
 
-		self.linear = torch.nn.Linear(n_filters+1, 1)
+	def forward(self, X, X_ctl=None):
+		"""A forward pass of the model.
+
+		This method takes in a nucleotide sequence X, a corresponding
+		per-position value from a control track, and a per-locus value
+		from the control track and makes predictions for the profile 
+		and for the counts. This per-locus value is usually the
+		log(sum(X_ctl_profile)+1) when the control is an experimental
+		read track but can also be the output from another model.
+
+		Parameters
+		----------
+		X: torch.tensor, shape=(batch_size, 4, sequence_length)
+			The one-hot encoded batch of sequences.
+
+		X_ctl: torch.tensor, shape=(batch_size, n_strands, sequence_length)
+			A value representing the signal of the control at each position in the
+			sequence.
+
+		Returns
+		-------
+		y_profile: torch.tensor, shape=(batch_size, n_strands, out_length)
+			The output predictions for each strand.
+		"""
 
-	def forward(self, X, X_ctl):
 		start, end = self.trimming, X.shape[2] - self.trimming
 
-		X = self.relu(self.iconv(X))
+		X = torch.nn.ReLU()(self.iconv(X))
 		for i in range(self.n_layers):
-			X_conv = self.relu(self.rconvs[i](X))
+			X_conv = torch.nn.ReLU()(self.rconvs[i](X))
 			X = torch.add(X, X_conv)
 
-		X = X[:, :, start:end]
+		if X_ctl is None:
+			X_w_ctl = X
+		else:
+			X_w_ctl = torch.cat([X, X_ctl], dim=1)
 
-		X_ = torch.cat([X, X_ctl], dim=1)
-		y_profile = self.fconv(X_)
-		y_profile = y_profile.reshape(X.shape[0], -1)
-		y_profile = self.logsoftmax(y_profile)
-		y_profile = y_profile.reshape(X.shape[0], self.n_outputs, -1)
+		y_profile = self.fconv(X_w_ctl)[:, :, start:end]
 
 		# counts prediction
-		X_ctl = torch.sum(X_ctl, axis=(1, 2)).reshape(-1, 1)
+		X = torch.mean(X[:, :, start-37:end+37], axis=2)
 
-		X = torch.mean(X, axis=2)
-		X = torch.cat([X, torch.log(X_ctl+1)], dim=-1)
+		if X_ctl is not None:
+			X_ctl = torch.sum(X_ctl[:, :, start-37:end+37], axis=(1, 2))
+			X_ctl = X_ctl.unsqueeze(-1)
+			X = torch.cat([X, torch.log(X_ctl+1)], dim=-1)
 
 		y_counts = self.linear(X).reshape(X.shape[0], 1)
 		return y_profile, y_counts
 
-	def predict(self, X, X_ctl, batch_size=64):
+	def predict(self, X, X_ctl=None, batch_size=64):
 		with torch.no_grad():
 			starts = numpy.arange(0, X.shape[0], batch_size)
 			ends = starts + batch_size
 
 			y_profiles, y_counts = [], []
 			for start, end in zip(starts, ends):
-				X_batch = X[start:end]
-				X_ctl_batch = X_ctl[start:end]
+				X_batch = X[start:end].cuda()
+				X_ctl_batch = None if X_ctl is None else X_ctl[start:end].cuda()
 
 				y_profiles_, y_counts_ = self(X_batch, X_ctl_batch)
+				y_profiles_ = y_profiles_.cpu()
+				y_counts_ = y_counts_.cpu()
 				
-				y_profiles.append(y_profiles_.cpu().detach().numpy())
-				y_counts.append(y_counts_.cpu().detach().numpy())
+				y_profiles.append(y_profiles_)
+				y_counts.append(y_counts_)
 
-			y_profiles = numpy.concatenate(y_profiles)
-			y_counts = numpy.concatenate(y_counts)
+			y_profiles = torch.cat(y_profiles)
+			y_counts = torch.cat(y_counts)
 			return y_profiles, y_counts
 
 	def fit_generator(self, training_data, optimizer, X_valid=None, 
@@ -163,13 +190,12 @@ class BPNet(torch.nn.Module):
 		validation_iter=100, verbose=True):
 
 		if X_valid is not None:
-			X_valid = torch.tensor(X_valid, dtype=torch.float32).cuda()
-			X_ctl_valid = torch.tensor(X_ctl_valid, dtype=torch.float32).cuda()
-
-			y_valid = y_valid.reshape(y_valid.shape[0], -1)
-			y_valid = numpy.expand_dims(y_valid, (1, 3))
+			X_valid = X_valid.cuda()
 			y_valid_counts = y_valid.sum(axis=2)
 
+		if X_ctl_valid is not None:
+			X_ctl_valid = X_ctl_valid.cuda()
+
 		columns = "Epoch\tIteration\tTraining Time\tValidation Time\t"
 		columns += "T MNLL\tT Count log1pMSE\t"
 		columns += "V MNLL\tV Profile Pearson\tV Count Pearson\tV Count log1pMSE"
@@ -177,15 +203,20 @@ class BPNet(torch.nn.Module):
 		if verbose:
 			print(columns)
 
-		start = time.time()
 		iteration = 0
 		best_loss = float("inf")
 
 		for epoch in range(max_epochs):
 			tic = time.time()
 
-			for X, X_ctl, y in training_data:
-				X, X_ctl, y = X.cuda(), X_ctl.cuda(), y.cuda()
+			for data in training_data:
+				if len(data) == 3:
+					X, X_ctl, y = data
+					X, X_ctl, y = X.cuda(), X_ctl.cuda(), y.cuda()
+				else:
+					X, y = data
+					X, y = X.cuda(), y.cuda()
+					X_ctl = None
 
 				# Clear the optimizer and set the model to training mode
 				optimizer.zero_grad()
@@ -193,10 +224,14 @@ class BPNet(torch.nn.Module):
 
 				# Run forward pass
 				y_profile, y_counts = self(X, X_ctl)
+				y_profile = y_profile.reshape(y_profile.shape[0], -1)
+				y_profile = torch.nn.functional.log_softmax(y_profile, dim=-1)
+				
+				y = y.reshape(y.shape[0], -1)
 
 				# Calculate the profile and count losses
-				profile_loss = MNLLLoss(y_profile, y)
-				count_loss = log1pMSELoss(y_counts, y.sum(dim=(1, 2)).reshape(-1, 1))
+				profile_loss = MNLLLoss(y_profile, y).mean()
+				count_loss = log1pMSELoss(y_counts, y.sum(dim=-1).reshape(-1, 1)).mean()
 
 				# Extract the profile loss for logging
 				profile_loss_ = profile_loss.item()
@@ -209,7 +244,7 @@ class BPNet(torch.nn.Module):
 
 				# Report measures if desired
 				if verbose and iteration % validation_iter == 0:
-					train_time = time.time() - start
+					train_time = time.time() - tic
 
 					with torch.no_grad():
 						self.eval()
@@ -218,12 +253,15 @@ class BPNet(torch.nn.Module):
 						y_profile, y_counts = self.predict(X_valid, X_ctl_valid)
 						valid_time = time.time() - tic
 
+						z = y_profile.shape
 						y_profile = y_profile.reshape(y_profile.shape[0], -1)
-						y_profile = numpy.expand_dims(y_profile, (1, 3))
-						y_counts = numpy.expand_dims(y_counts, 1)
+						y_profile = torch.nn.functional.log_softmax(y_profile, dim=-1)
+						y_profile = y_profile.reshape(*z)
 
-						measures = compute_performance_metrics(y_valid, y_profile, 
-							y_valid_counts, y_counts, 7, 81)
+						measures = calculate_performance_measures(y_profile, 
+							y_valid, y_counts, kernel_sigma=7, 
+							kernel_width=81, measures=['profile_mnll', 
+							'profile_pearson', 'count_pearson', 'count_mse'])
 
 						line = "{}\t{}\t{:4.4}\t{:4.4}\t".format(epoch, iteration,
 							train_time, valid_time)
@@ -232,24 +270,21 @@ class BPNet(torch.nn.Module):
 							count_loss_)
 
 						line += "{:4.4}\t{:4.4}\t{:4.4}\t{:4.4}".format(
-							measures['nll'].mean(), 
-							measures['profile_pearson'].mean(),
-							measures['count_pearson'].mean(), 
+							measures['profile_mnll'].mean(), 
+							numpy.nan_to_num(measures['profile_pearson']).mean(),
+							numpy.nan_to_num(measures['count_pearson']).mean(), 
 							measures['count_mse'].mean()
 						)
 
-						valid_loss = measures['nll'].mean() + self.alpha * measures['count_mse'].mean()
+						valid_loss = measures['profile_mnll'].mean() + self.alpha * measures['count_mse'].mean()
 						line += "\t{}".format(valid_loss < best_loss)
 
+						if valid_loss < best_loss:
+							torch.save(self, "{}.torch".format(self.name))
+							best_loss = valid_loss
+					
 						print(line)
 
-						start = time.time()
-
-						if valid_loss < best_loss:
-							best_loss = valid_loss
-
-							self = self.cpu()
-							torch.save(self, "bpnet.{}.{}.torch".format(self.n_filters, self.n_layers))
-							self = self.cuda()
-					
 				iteration += 1
+
+		torch.save(self, "{}.final.torch".format(self.name))
