@@ -36,13 +36,10 @@ def MNLLLoss(logps, true_counts):
 		dimensions.
 	"""
 
-	logps = logps.reshape(logps.shape[0], -1)
-	true_counts = true_counts.reshape(true_counts.shape[0], -1)
-
 	log_fact_sum = torch.lgamma(torch.sum(true_counts, dim=-1) + 1)
 	log_prod_fact = torch.sum(torch.lgamma(true_counts + 1), dim=-1)
 	log_prod_exp = torch.sum(true_counts * logps, dim=-1)
-	return -torch.mean(log_fact_sum - log_prod_fact + log_prod_exp)
+	return -log_fact_sum + log_prod_fact - log_prod_exp
 
 def log1pMSELoss(log_predicted_counts, true_counts):
 	"""A MSE loss on the log(x+1) of the inputs.
@@ -67,10 +64,10 @@ def log1pMSELoss(log_predicted_counts, true_counts):
 
 	Returns
 	-------
-	loss: float
+	loss: torch.tensor, shape=(n, 1)
 		The MSE loss on the log of the two inputs, averaged over all examples
 		and all other dimensions.
 	"""
 
 	log_true = torch.log(true_counts+1)
-	return torch.nn.MSELoss()(log_predicted_counts, log_true)
\ No newline at end of file
+	return torch.mean(torch.square(log_true - log_predicted_counts), dim=-1)
